import collections
import re
import time

import spacy
from chatty_goose.settings import HQESettings
from pyserini.search import SimpleSearcher

from .cqr import CQR

__all__ = ["HQE"]

nlp = spacy.load("en_core_web_sm", parser=False, ner=False, textcat=False)
nlp.pipeline = [nlp.pipeline[0]]
STOP_WORDS = nlp.Defaults.stop_words


class HQE(CQR):
    """Historical Query Expansion for conversational query reformulation"""

    def __init__(self, searcher: SimpleSearcher, settings: HQESettings = HQESettings()):
        super().__init__("HQE", verbose=settings.verbose)

        # Model settings
        self.M = settings.M
        self.eta = settings.eta
        self.R_topic = settings.R_topic
        self.R_sub = settings.R_sub
        self.pos_filter = settings.filter
        self.searcher = searcher

        # History
        self.key_word_list = collections.defaultdict(list)
        self.subkey_word_list = collections.defaultdict(list)

    def rewrite(self, query: str) -> str:
        start_time = time.time()
        self.turn_id += 1
        self.key_word_extraction(query)
        if self.turn_id != 0:
            hits = self.searcher.search(query, 1)
            key_word = self.query_expansion(self.key_word_list, 0, self.turn_id)
            subkey_word = ""
            if len(hits) == 0 or hits[0].score <= self.eta:
                end_turn = self.turn_id + 1
                start_turn = end_turn - self.M
                if start_turn < 0:
                    start_turn = 0
                subkey_word = self.query_expansion(
                    self.subkey_word_list, start_turn, end_turn
                )
            query = key_word + " " + subkey_word + " " + query
        self.total_latency += time.time() - start_time
        return query

    def reset_history(self):
        super().reset_history()
        self.key_word_list = collections.defaultdict(list)
        self.subkey_word_list = collections.defaultdict(list)

    def key_word_extraction(self, query):
        proc_query = self.calc_word_score(query)
        # Extract topic keyword
        if self.pos_filter == "no":
            for i, word in enumerate(proc_query["word"]):
                if proc_query["score"][i] >= self.R_topic:
                    self.key_word_list[self.turn_id].append(word)
                if (proc_query["score"][i] >= self.R_sub) & (
                    proc_query["score"][i] < self.R_topic
                ):
                    self.subkey_word_list[self.turn_id].append(word)
        elif self.pos_filter == "pos":
            for i, word in enumerate(proc_query["word"]):
                if ("NN" in proc_query["pos"][i]) or ("JJ" in proc_query["pos"][i]):
                    if proc_query["score"][i] >= self.R_topic:
                        self.key_word_list[self.turn_id].append(word)
                    if (proc_query["score"][i] >= self.R_sub) & (
                        proc_query["score"][i] < self.R_topic
                    ):
                        self.subkey_word_list[self.turn_id].append(word)
        elif self.pos_filter == "stp":
            for i, word in enumerate(proc_query["word"]):
                if word not in STOP_WORDS:
                    if proc_query["score"][i] >= self.R_topic:
                        self.key_word_list[self.turn_id].append(word)
                    if (proc_query["score"][i] >= self.R_sub) & (
                        proc_query["score"][i] < self.R_topic
                    ):
                        self.subkey_word_list[self.turn_id].append(word)

    def calc_word_score(self, query):
        nlp_query = nlp(pre_process(query))
        proc_query = process(nlp_query)
        query_words = proc_query["word"]
        proc_query["score"] = []

        for query_word in query_words:
            hits = self.searcher.search(query_word, 1)
            try:
                score = hits[0].score
                proc_query["score"].append(score)
            except:
                proc_query["score"].append(-1)

        return proc_query

    @staticmethod
    def query_expansion(key_word_list, start_turn, end_turn):
        query_expansion = ""
        for turn in range(start_turn, end_turn + 1):
            for word in key_word_list[turn]:
                query_expansion = query_expansion + " " + word
        return query_expansion


def pre_process(text):
    text = re.sub(
        u"-|\u2010|\u2011|\u2012|\u2013|\u2014|\u2015|%|\[|\]|:|\(|\)|/|\t",
        _space_extend,
        text,
    )
    text = text.strip(" \n")
    text = re.sub("\s+", " ", text)
    return text


def process(parsed_text):
    output = {
        "word": [],
        "lemma": [],
        "pos": [],
        "pos_id": [],
        "ent": [],
        "ent_id": [],
        "offsets": [],
        "sentences": [],
    }

    for token in parsed_text:
        output["word"].append(_str(token.text))
        pos = token.tag_
        output["pos"].append(pos)

    return output


def _space_extend(matchobj):
    return " " + matchobj.group(0) + " "


def _str(s):
    """ Convert PTB tokens to normal tokens """
    if s.lower() == "-lrb-":
        s = "("
    elif s.lower() == "-rrb-":
        s = ")"
    elif s.lower() == "-lsb-":
        s = "["
    elif s.lower() == "-rsb-":
        s = "]"
    elif s.lower() == "-lcb-":
        s = "{"
    elif s.lower() == "-rcb-":
        s = "}"
    return s



from pydantic import BaseSettings

from chatty_goose.types import PosFilter

__all__ = ["SearcherSettings", "HQESettings", "T5Settings"]


class SearcherSettings(BaseSettings):
    """Settings for Anserini searcher"""

    index_path: str  # Pre-built index name or path to Lucene index 
    k1: float = 0.82  # BM25 k parameter
    b: float = 0.68  # BM25 b parameter
    rm3: bool = False  # use RM3
    fb_terms: int = 10  # RM3 number of expansion trees
    fb_docs: int = 10  # RM3 number of documents
    original_query_weight: float = 0.8  # RM3 weigh to assign initial query


class CQRSettings(BaseSettings):
    verbose: bool = False


class HQESettings(CQRSettings):
    """Settings for HQE with defaults tuned on CAsT"""

    M: int = 5  # number of aggregate historical queries
    eta: float = 10.0  # QPP threshold for first stage retrieval
    R_topic: float = 4.5  # topic keyword threshold
    R_sub: float = 3.5  # subtopic keyword threshold
    filter: PosFilter = PosFilter.POS  # 'no' or 'pos' or 'stp'


class T5Settings(CQRSettings):
    """Settings for T5 model for NTR"""

    model_name: str = "castorini/t5-base-canard"
    max_length: int = 64
    num_beams: int = 10
    early_stopping: bool = True



import logging
from os import path
from typing import Dict, List, Tuple

from pygaggle.rerank.transformer import MonoBERT
from pyserini.search import JSimpleSearcherResult, SimpleSearcher

from chatty_goose.settings import SearcherSettings


def reciprocal_rank_fusion(
    hit_lists: List[List[JSimpleSearcherResult]], k: int = 60
) -> List[JSimpleSearcherResult]:
    """
    Implements reciprocal rank fusion as defined in
    "Reciprocal Rank Fusion Outperforms Condorcet and Individual Rank Learning Methods" by Cormack, Clarke and Buettcher.
    
    Parameters:
        hit_lists: lists of hits to merge using reciprocal rank fusion
        k: term to avoid vanishing importance of lower-ranked documents (default 60 from original paper)
    """
    if len(hit_lists) == 0:
        return []

    if len(hit_lists) == 1:
        return hit_lists[0]

    doc_scores: Dict[str, Tuple[float, JSimpleSearcherResult]] = {}
    for hits in hit_lists:
        for pos, hit in enumerate(hits, start=1):
            cur_rank = doc_scores.get(hit.docid, (0.0, hit))[0]
            doc_scores[hit.docid] = (cur_rank + 1.0 / (k + pos), hit)

    # Sort by highest score
    result = [
        score_hit[1]
        for _, score_hit in sorted(iter(doc_scores.items()), key=lambda kv: -kv[1][0])
    ]
    return result


def build_bert_reranker(
    name_or_path: str = "castorini/monobert-large-msmarco-finetune-only",
    device: str = None,
):
    """Returns a BERT reranker using the provided model name or path to load from"""
    model = MonoBERT.get_model(name_or_path, device=device)
    tokenizer = MonoBERT.get_tokenizer(name_or_path)
    return MonoBERT(model, tokenizer)


def build_searcher(settings: SearcherSettings) -> SimpleSearcher:
    if path.isdir(settings.index_path):
        searcher = SimpleSearcher(settings.index_path)
    else:
        searcher = SimpleSearcher.from_prebuilt_index(settings.index_path)
    searcher.set_bm25(float(settings.k1), float(settings.b))
    logging.info(
        "Initializing BM25, setting k1={} and b={}".format(settings.k1, settings.b)
    )
    if settings.rm3:
        searcher.set_rm3(
            settings.fb_terms, settings.fb_docs, settings.original_query_weight
        )
        logging.info(
            "Initializing RM3, setting fbTerms={}, fbDocs={} and originalQueryWeight={}".format(
                settings.fb_terms, settings.fb_docs, settings.original_query_weight
            )
        )
    return searcher
